"""
The `clintest.outcome.Outcome` of a test, accessible via `clintest.test.Test.outcome()`.

The outcome of a test may be either

- `F?` (possibly false),
- `T?` (possibly true),
- `F!` (certainly false), or
- `T!` (certainly true).

These four options are represented using two booleans within `Outcome`:

- `current_value` stores the actual result of the test.
- `is_certain` indicates whether the result is certain.
"""

from typing import Tuple

class Outcome:
    """
    The outcome of a test.

    Parameters
    ----------
    current_value
        The actual result of the test.
    is_certain
        Whether the `current_value` is certain.
    """

    def __init__(self, current_value: bool, is_certain: bool) -> None:
        self.__current_value = current_value
        self.__is_certain = is_certain

    def __repr__(self):
        name = self.__class__.__name__
        return f"{name}({self.__current_value}, {self.__is_certain})"

    def __str__(self):
        return str(self.__current_value)[:1] + ["?", "!"][self.__is_certain]

    def __eq__(self, other):
        # pylint: disable=protected-access
        return self.__current_value == other.__current_value \
            and self.__is_certain == other.__is_certain

    def current_value(self) -> bool:
        """
        Returns the `current_value` of this outcome.
        """

        return self.__current_value

    def is_certain(self) -> bool:
        """
        Returns whether this outcome `is_certain`.
        """

        return self.__is_certain

    def as_tuple(self) -> Tuple[bool, bool]:
        """
        Returns this outcome as a tuple `(current_value, is_certain)`.
        """

        return (self.__current_value, self.__is_certain)

    def is_certainly_true(self) -> bool:
        """
        Returns whether this outcome is certainly true, i.e., if `is_certain and current_value` holds.
        """

        return self.__is_certain and self.__current_value

    def is_certainly_false(self) -> bool:
        """
        Returns whether this outcome is certainly false, i.e., if `is_certain and not current_value` holds.
        """

        return self.__is_certain and not self.__current_value



# pylint: disable=import-outside-toplevel
# pylint: disable=missing-module-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring


def test_false_uncertain():
    from clintest.outcome import Outcome
    outcome = Outcome(False, False)

    assert not outcome.current_value()
    assert not outcome.is_certain()
    assert     outcome.as_tuple() == (False, False)
    assert not outcome.is_certainly_false()
    assert not outcome.is_certainly_true()


def test_false_certain():
    from clintest.outcome import Outcome
    outcome = Outcome(False, True)

    assert not outcome.current_value()
    assert     outcome.is_certain()
    assert     outcome.as_tuple() == (False, True)
    assert     outcome.is_certainly_false()
    assert not outcome.is_certainly_true()


def test_true_uncertain():
    from clintest.outcome import Outcome
    outcome = Outcome(True, False)

    assert     outcome.current_value()
    assert not outcome.is_certain()
    assert     outcome.as_tuple() == (True, False)
    assert not outcome.is_certainly_false()
    assert not outcome.is_certainly_true()


def test_true_certain():
    from clintest.outcome import Outcome
    outcome = Outcome(True, True)

    assert     outcome.current_value()
    assert     outcome.is_certain()
    assert     outcome.as_tuple() == (True, True)
    assert not outcome.is_certainly_false()
    assert     outcome.is_certainly_true()



"""
A test framework for `clingo` programs.

`clintest` is a test framework written in Python that makes it easy to write efficient tests for `clingo` programs.
It provides you with numerous off-the-shelf components that allow you to assemble the most commonly used tests quickly, saving you the time to write them yourself.
However, should you require a custom-build test, it will work along the others just fine.

In order to avoid time wasted on unnecessary computations, `clintest` will monitor the outcome of your test while steering the solving process.
Once the outcome of your test is certain, it will automatically tell the solver to abort the search for further solutions.

As `clintest` is focussed on the specifics of `clingo` programs, it works best if you combine it with a general purpose frameworks like `pytest`.

## Installation
This framework is guaranteed to work with Python 3.8 or greater.
You have several options to install it:

### Using conda
A conda package is planned but currently not available.

### Using pip
The pip package is hosted at https://pypi.org/project/clintest.
It can be installed with

```
$ pip install clintest
```

### From source
The project is hosted on GitHub at https://github.com/potassco/clintest and can also be installed from source. We recommend this only for development purposes.

```
$ git clone https://github.com/potassco/clintest
$ cd clintest
$ pip install -e .
```

## Usage
This section is meant to guide you through the most important features of `clintest` using simple examples.

### Inspecting models
Imagine you have written the program `a. {b}.` and want to ensure that all its models contains the atom `a`.
In order to do so, you first need to create a `test.Test`.

```
from clintest.test import Assert
from clintest.quantifier import All
from clintest.assertion import Contains

test = Assert(All(), Contains("a"))
```

The test `test.Assert` inspects the models of a program.
It needs to be initialized with an `assertion.Assertion` and a `quantifier.Quantifier`.
An assertion (here: `assertion.Contains`) is a statement that may or may not hold for a certain model.
A quantifier (here: `quantifier.All`) specifies how many assertions must hold in order to pass the test.

Tests retain their own outcome which may be either

  - `F?` (possibly false),
  - `T?` (possibly true),
  - `F!` (certainly false), or
  - `T!` (certainly true).

Having an uncertain outcome means that a test has not (yet) been completed.
Once the outcome is certain, it must no longer change in order for `clintest` to function properly.
The outcome of a test can be queried as follows.

>>> print(test.outcome())
T?

Due to the fact this test was never run, its outcome is still uncertain.
In order to run the test, we need to create a `solver.Solver` first.

```
from clintest.solver import Clingo
solver = Clingo("0", "a. {b}.")
```

The solver `solver.Clingo` is a facade around `clingo.control.Control`.
The constructor expects the program (here: `"a. {b}."`) and a list of arguments for the solver (here: `"0"`, meaning that the solver should compute all models).
Once a solver is set up, it may solve your test as follows.

>>> solver.solve(test)
>>> print(test.outcome())
T!

As you realize from the output, any model of `a. {b}.` does indeed contain the atom `a`.
If you want to ensure this within a framework like `pytest`, use `clintest.test.Test.assert_` to raise an `AssertionError` with a proper message if the test's outcome is not certainly true (`T!`).

```
test.assert_()
```

### Compound tests
Testing real-world programs would still require a lot of boilerplate, if `clintest` had no support for combining simple tests into more complex ones.
The following example illustrates how to build a test that simultaneously ensures

  - any model contains the atom `a`,
  - all models contain the atom `b`, and
  - any model contains the atom `c`.

```
from clintest.test import Assert, And
from clintest.quantifier import All, Any
from clintest.assertion import Contains

test = And(
    Assert(Any(), Contains("a")),
    Assert(All(), Contains("b")),
    Assert(Any(), Contains("c")),
)
```

Solving this test with our previous solver leads to the following result.

>>> solver.solve(test)
>>> test.assert_()
Traceback (most recent call last):
  File "<stdin>", line 1, in <module>
  File "clintest/clintest/test.py", line 42, in assert_
    raise AssertionError(msg)
AssertionError: The following test has failed.
    [F!] And
        operands:
             0: [T!] Assert
                quantifier: Any
                assertion:  Contains("a")
             1: [F!] Assert
                quantifier: All
                assertion:  Contains("b")
             2: [F?] Assert
                quantifier: Any
                assertion:  Contains("c")
        short_circuit:  True
        ignore_certain: True

Because the test has failed, `clintest.test.Test.assert_` produces a rather detailed exception which may be used to explain the cause of the failure:
Test #1 did fail, probably because there was a model that did not contain the atom `b`.

But if you watch carefully, there is something else to discover.
Test #2 was never completed.
This is due to a deliberate optimization as the outcome of test #2 was irrelevant once the outcome of test #1 was certainly false (`F!`).

### Debugging
Understanding why a test has failed can be challenging if one has no insight into the interaction between the test and the solver.
This is where `clintest.test.Record`, a test that can be wrapped around any other test, comes in handy.

```
from clintest.test import Record
record = Record(test)
```

To the solver or the surrounding compound test, `record` behaves just as <code>test</code>.
Any call to one of its `on_*`-methods is forwarded to the respective method of <code>test</code>.
The only difference is that `record` also keeps a detailed `test.Recording`of these function calls.
Using the <code>test</code> from the previous section, the following example is what a recording may look like.
In order to obtain the same result, make sure <code>test</code> was not solved before as a previously solved test cannot be solved again and therefore leads to a different recording.

>>> solver.solve(record)
>>> print(record)
[F!] Record
    test: [F!] And
        operands:
             0: [T!] Assert
                quantifier: Any
                assertion:  Contains("a")
             1: [F!] Assert
                quantifier: All
                assertion:  Contains("b")
             2: [F?] Assert
                quantifier: Any
                assertion:  Contains("c")
        short_circuit:  True
        ignore_certain: True
    recording:
        0: [T?] __init__
        1: [F!] on_model
            a
        2: [F!] on_statistics
        3: [F!] on_finish

From this recording we learn that the absence of atom `b` in a model was indeed the reason for the failure of the test.

### Custom-build tests
In case you are unable to assemble your test from the off-the-shelf components in `clintest`, you might consider to custom-build it.
Custom-build tests must, just as any other tests, extend `test.Test` in order to work with this library.
This includes implementing two methods though a third is often needed:

1. `test.Test.outcome()` may me called anytime and should return the current `outcome.Outcome` of your test.
Once the outcome is certain, it must not change anymore.
2. `test.Test.on_finish()` is called once solving comes to an end.
The call to this method is your last change to alter the outcome of your test.
After the call, the outcome must be certain.
3. optional: `test.Test.on_model()` is called whenever a model is found.
You may inspect the model to change the outcome of your test.
This method should return if additional models are necessary to decide the test.
In case it returns `False`, the outcome of the test must be certain.

There are further `on_*`-methods you may override.
For detailed information, refer to the class level documentation of `test.Test`.
You may also draw inspiration from the tests implemented in `test`.

Instead of custom-building a whole test, it often advisable to implement the missing components.
This approach is often less labor-intensive as it aligns more seamlessly with the modular approach of `clintest`.
See `quantifier.Quantifier` and `assertion.Assertion`.

If you have an idea for an addition to `clintest` that could benefit other users as well, we encourage you to submit a pull request or open an issue.
We are happy to consider including your idea into this project.
"""



# pylint: disable=import-outside-toplevel
# pylint: disable=missing-module-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring
# pylint: disable=redefined-outer-name


import pytest


@pytest.fixture
def solver():
    from clintest.solver import Clingo
    return Clingo("0", "a. {b}.")


@pytest.fixture
def recording_no_model():
    from clintest.test import Recording
    return Recording([
        {'__f': '__init__'},
    ])


@pytest.fixture
def recording_one_model():
    from clintest.test import Recording
    return Recording([
        {'__f': '__init__'},
        {'__f': 'on_model', 'str(model)': 'a'},
        {'__f': 'on_statistics'},
        {'__f': 'on_finish'},
    ])


@pytest.fixture
def recording_two_models():
    from clintest.test import Recording
    return Recording([
        {'__f': '__init__'},
        {'__f': 'on_model', 'str(model)': 'a'},
        {'__f': 'on_model', 'str(model)': 'b a'},
        {'__f': 'on_statistics'},
        {'__f': 'on_finish'},
    ])


def test_assert_all(solver, recording_one_model, recording_two_models):
    from clintest.test import Assert, Record
    from clintest.quantifier import All
    from clintest.assertion import Contains

    test = Record(Assert(All(), Contains("a")))
    solver.solve(test)
    assert test.outcome().is_certainly_true()
    assert recording_two_models.subsumes(test.recording)

    test = Record(Assert(All(), Contains("b")))
    solver.solve(test)
    assert test.outcome().is_certainly_false()
    assert recording_one_model.subsumes(test.recording)


def test_assert_any(solver, recording_one_model, recording_two_models):
    from clintest.test import Assert, Record
    from clintest.quantifier import Any
    from clintest.assertion import Contains

    test = Record(Assert(Any(), Contains("a")))
    solver.solve(test)
    assert test.outcome().is_certainly_true()
    assert recording_one_model.subsumes(test.recording)

    test = Record(Assert(Any(), Contains("b")))
    solver.solve(test)
    assert test.outcome().is_certainly_true()
    assert recording_two_models.subsumes(test.recording)

    test = Record(Assert(Any(), Contains("c")))
    solver.solve(test)
    assert test.outcome().is_certainly_false()
    assert recording_two_models.subsumes(test.recording)


def test_assert_exact(solver, recording_one_model, recording_two_models):
    from clintest.test import Assert, Record
    from clintest.quantifier import Exact
    from clintest.assertion import Contains

    test = Record(Assert(Exact(0), Contains("a")))
    solver.solve(test)
    assert test.outcome().is_certainly_false()
    assert recording_one_model.subsumes(test.recording)

    test = Record(Assert(Exact(1), Contains("a")))
    solver.solve(test)
    assert test.outcome().is_certainly_false()
    recording_two_models.subsumes(test.recording)

    test = Record(Assert(Exact(2), Contains("a")))
    solver.solve(test)
    assert test.outcome().is_certainly_true()
    recording_two_models.subsumes(test.recording)

    test = Record(Assert(Exact(0), Contains("b")))
    solver.solve(test)
    assert test.outcome().is_certainly_false()
    recording_two_models.subsumes(test.recording)

    test = Record(Assert(Exact(1), Contains("b")))
    solver.solve(test)
    assert test.outcome().is_certainly_true()
    recording_two_models.subsumes(test.recording)

    test = Record(Assert(Exact(2), Contains("b")))
    solver.solve(test)
    assert test.outcome().is_certainly_false()
    recording_two_models.subsumes(test.recording)


def test_true(solver, recording_no_model, recording_two_models):
    from clintest.test import True_, Record

    test = Record(True_())
    solver.solve(test)
    assert test.outcome().is_certainly_true()
    assert recording_no_model.subsumes(test.recording)

    test = Record(True_(lazy = False))
    solver.solve(test)
    assert test.outcome().is_certainly_true()
    assert recording_two_models.subsumes(test.recording)


def test_false(solver, recording_no_model, recording_two_models):
    from clintest.test import False_, Record

    test = Record(False_())
    solver.solve(test)
    assert test.outcome().is_certainly_false()
    assert recording_no_model.subsumes(test.recording)

    test = Record(False_(lazy = False))
    solver.solve(test)
    assert test.outcome().is_certainly_false()
    assert recording_two_models.subsumes(test.recording)


def test_not(solver, recording_no_model):
    from clintest.test import False_, True_, Not, Record

    inner = Record(False_())
    outer = Record(Not(inner))
    solver.solve(outer)
    assert outer.outcome().is_certainly_true()
    assert recording_no_model.subsumes(inner.recording)
    assert recording_no_model.subsumes(outer.recording)

    inner = Record(True_())
    outer = Record(Not(inner))
    solver.solve(outer)
    assert outer.outcome().is_certainly_false()
    assert recording_no_model.subsumes(inner.recording)
    assert recording_no_model.subsumes(outer.recording)


def test_and(solver, recording_no_model):
    from clintest.test import False_, True_, And, Record

    inner = [Record(test) for test in [False_(), False_()]]
    outer = Record(And(*inner))
    solver.solve(outer)
    assert outer.outcome().is_certainly_false()
    assert recording_no_model.subsumes(outer.recording)
    assert recording_no_model.subsumes(inner[0].recording)
    assert recording_no_model.subsumes(inner[1].recording)

    inner = [Record(test) for test in [False_(), True_()]]
    outer = Record(And(*inner))
    solver.solve(outer)
    assert outer.outcome().is_certainly_false()
    assert recording_no_model.subsumes(outer.recording)
    assert recording_no_model.subsumes(inner[0].recording)
    assert recording_no_model.subsumes(inner[1].recording)

    inner = [Record(test) for test in [True_(), False_()]]
    outer = Record(And(*inner))
    solver.solve(outer)
    assert outer.outcome().is_certainly_false()
    assert recording_no_model.subsumes(outer.recording)
    assert recording_no_model.subsumes(inner[0].recording)
    assert recording_no_model.subsumes(inner[1].recording)

    inner = [Record(test) for test in [True_(), True_()]]
    outer = Record(And(*inner))
    solver.solve(outer)
    assert outer.outcome().is_certainly_true()
    assert recording_no_model.subsumes(outer.recording)
    assert recording_no_model.subsumes(inner[0].recording)
    assert recording_no_model.subsumes(inner[1].recording)


def test_and_ignore_certain(solver, recording_two_models):
    from clintest.test import And, Assert, Record, Recording
    from clintest.assertion import Contains
    from clintest.quantifier import Any

    inner = [Record(test) for test in [
        Assert(Any(), Contains("a")),
        Assert(Any(), Contains("b"))
    ]]
    outer = Record(And(*inner))
    solver.solve(outer)
    assert outer.outcome().is_certainly_true()
    assert recording_two_models.subsumes(outer.recording)
    assert Recording([
        {'__f': '__init__'},
        {'__f': 'on_model', 'str(model)': 'a'},
    ]).subsumes(inner[0].recording)
    assert Recording([
        {'__f': '__init__'},
        {'__f': 'on_model', 'str(model)': 'a'},
        {'__f': 'on_model', 'str(model)': 'b a'},
    ]).subsumes(inner[1].recording)

    inner = [Record(test) for test in [
        Assert(Any(), Contains("a")),
        Assert(Any(), Contains("b"))
    ]]
    outer = Record(And(*inner, ignore_certain=False))
    solver.solve(outer)
    assert outer.outcome().is_certainly_true()
    assert recording_two_models.subsumes(outer.recording)
    assert recording_two_models.subsumes(inner[0].recording)
    assert recording_two_models.subsumes(inner[1].recording)


def test_and_short_circuit(solver, recording_one_model, recording_two_models):
    from clintest.test import And, Assert, Record, Recording
    from clintest.assertion import Contains
    from clintest.quantifier import All

    inner = [Record(test) for test in [
        Assert(All(), Contains('b')),
        Assert(All(), Contains('a'))
    ]]
    outer = Record(And(*inner))
    solver.solve(outer)
    assert outer.outcome().is_certainly_false()
    assert recording_one_model.subsumes(outer.recording)
    assert Recording([
        {'__f': '__init__'},
        {'__f': 'on_model', 'str(model)': 'a'}
    ]).subsumes(inner[0].recording)
    assert Recording([
        {'__f': '__init__'},
    ]).subsumes(inner[1].recording)

    inner = [Record(test) for test in [
        Assert(All(), Contains('b')),
        Assert(All(), Contains('a'))
    ]]
    outer = Record(And(*inner, short_circuit=False))
    solver.solve(outer)
    assert outer.outcome().is_certainly_false()
    assert recording_two_models.subsumes(outer.recording)
    assert Recording([
        {'__f': '__init__'},
        {'__f': 'on_model', 'str(model)': 'a'}
    ]).subsumes(inner[0].recording)
    assert recording_two_models.subsumes(inner[1].recording)


def test_or(solver, recording_no_model):
    from clintest.test import False_, True_, Or, Record

    inner = [Record(test) for test in [False_(), False_()]]
    outer = Record(Or(*inner))
    solver.solve(outer)
    assert outer.outcome().is_certainly_false()
    assert recording_no_model.subsumes(outer.recording)
    assert recording_no_model.subsumes(inner[0].recording)
    assert recording_no_model.subsumes(inner[1].recording)

    inner = [Record(test) for test in [False_(), True_()]]
    outer = Record(Or(*inner))
    solver.solve(outer)
    assert outer.outcome().is_certainly_true()
    assert recording_no_model.subsumes(outer.recording)
    assert recording_no_model.subsumes(inner[0].recording)
    assert recording_no_model.subsumes(inner[1].recording)

    inner = [Record(test) for test in [True_(), False_()]]
    outer = Record(Or(*inner))
    solver.solve(outer)
    assert outer.outcome().is_certainly_true()
    assert recording_no_model.subsumes(outer.recording)
    assert recording_no_model.subsumes(inner[0].recording)
    assert recording_no_model.subsumes(inner[1].recording)

    inner = [Record(test) for test in [True_(), True_()]]
    outer = Record(Or(*inner))
    solver.solve(outer)
    assert outer.outcome().is_certainly_true()
    assert recording_no_model.subsumes(outer.recording)
    assert recording_no_model.subsumes(inner[0].recording)
    assert recording_no_model.subsumes(inner[1].recording)


def test_or_ignore_certain(solver, recording_two_models):
    from clintest.test import Assert, Not, Or, Record, Recording
    from clintest.assertion import Contains
    from clintest.quantifier import Any

    inner = [Record(test) for test in [
        Not(Assert(Any(), Contains("a"))),
        Not(Assert(Any(), Contains("b")))
    ]]
    outer = Record(Or(*inner))
    solver.solve(outer)
    assert outer.outcome().is_certainly_false()
    assert recording_two_models.subsumes(outer.recording)
    assert Recording([
        {'__f': '__init__'},
        {'__f': 'on_model', 'str(model)': 'a'},
    ]).subsumes(inner[0].recording)
    assert Recording([
        {'__f': '__init__'},
        {'__f': 'on_model', 'str(model)': 'a'},
        {'__f': 'on_model', 'str(model)': 'b a'},
    ]).subsumes(inner[1].recording)

    inner = [Record(test) for test in [
        Not(Assert(Any(), Contains("a"))),
        Not(Assert(Any(), Contains("b")))
    ]]
    outer = Record(Or(*inner, ignore_certain=False))
    solver.solve(outer)
    assert outer.outcome().is_certainly_false()
    assert recording_two_models.subsumes(outer.recording)
    assert recording_two_models.subsumes(inner[0].recording)
    assert recording_two_models.subsumes(inner[1].recording)


def test_or_short_circuit(solver, recording_one_model, recording_two_models):
    from clintest.test import Assert, Or, Not, Record, Recording
    from clintest.assertion import Contains
    from clintest.quantifier import All

    inner = [Record(test) for test in [
        Not(Assert(All(), Contains('b'))),
        Not(Assert(All(), Contains('a')))
    ]]
    outer = Record(Or(*inner))
    solver.solve(outer)
    assert outer.outcome().is_certainly_true()
    assert recording_one_model.subsumes(outer.recording)
    assert Recording([
        {'__f': '__init__'},
        {'__f': 'on_model', 'str(model)': 'a'}
    ]).subsumes(inner[0].recording)
    assert Recording([
        {'__f': '__init__'},
    ]).subsumes(inner[1].recording)

    inner = [Record(test) for test in [
        Not(Assert(All(), Contains('b'))),
        Not(Assert(All(), Contains('a')))
    ]]
    outer = Record(Or(*inner, short_circuit=False))
    solver.solve(outer)
    assert outer.outcome().is_certainly_true()
    assert recording_two_models.subsumes(outer.recording)
    assert Recording([
        {'__f': '__init__'},
        {'__f': 'on_model', 'str(model)': 'a'}
    ]).subsumes(inner[0].recording)
    assert recording_two_models.subsumes(inner[1].recording)



"""
The abstract class `clintest.assertion.Assertion` and classes extending it.
"""

from abc import ABC, abstractmethod
from typing import Set, Union

from clingo.solving import Model
from clingo.symbol import Symbol, parse_term


def _into_symbol(symbol: Union[Symbol, str]) -> Symbol:
    if isinstance(symbol, Symbol):
        return symbol
    return parse_term(symbol)


class Assertion(ABC):
    """
    An assertion is a statement that may or may not hold for a certain `clingo.model.Model`.
    As such, one is necessary to assemble the `clintest.test.Assert` test.
    """

    @abstractmethod
    def holds_for(self, model: Model) -> bool:
        """
        Returns whether this assertions holds for `model`.

        Parameters
        ----------
        model
            A `clingo.model.Model`.


        Returns
        -------
        Whether this assertions holds for `model`.
        """


class Contains(Assertion):
    """
    An assertion that holds if a model contains a given `symbol`.

    Parameters
    ----------
    symbol
        The `clingo.symbol.Symbol` or a `str` that can be parsed into a `clingo.symbol.Symbol` with `clingo.symbol.parse_term`.
    """

    def __init__(self, symbol: Union[Symbol, str]) -> None:
        self.__symbol = _into_symbol(symbol)

    def __repr__(self):
        name = self.__class__.__name__
        return f"{name}(\"{self.__symbol}\")"

    def holds_for(self, model: Model) -> bool:
        return model.contains(self.__symbol)


class Equals(Assertion):
    """
    An assertion that holds if the symbols of a model are equals to a given set of `symbols`.

    Parameters
    ----------
    symbols
        A set of `clingo.symbol.Symbol`s or `str`s that can be parsed into a `clingo.symbol.Symbol`s with `clingo.symbol.parse_term`.
    """

    def __init__(self, symbols: Set[Union[Symbol, str]]) -> None:
        self.__symbols = {_into_symbol(s) for s in symbols}

    def __repr__(self):
        name = self.__class__.__name__
        symbols = {str(symbol) for symbol in self.__symbols}
        return f"{name}({symbols})"

    def holds_for(self, model: Model) -> bool:
        return self.__symbols == set(model.symbols(shown=True))


class SubsetOf(Assertion):
    """
    An assertion that holds if the symbols of a model are a subset of a given set of `symbols`.

    Parameters
    ----------
    symbols
        A set of `clingo.symbol.Symbol`s or `str`s that can be parsed into a `clingo.symbol.Symbol`s with `clingo.symbol.parse_term`.
    """

    def __init__(self, symbols: Set[Union[Symbol, str]]) -> None:
        self.__symbols = {_into_symbol(s) for s in symbols}

    def __repr__(self):
        name = self.__class__.__name__
        symbols = {str(symbol) for symbol in self.__symbols}
        return f"{name}({symbols})"

    def holds_for(self, model: Model) -> bool:
        return set(model.symbols(shown=True)).issubset(self.__symbols)


class SupersetOf(Assertion):
    """
    An assertion that holds if the symbols of a model are a superset of a given set of `symbols`.

    Parameters
    ----------
    symbols
        A set of `clingo.symbol.Symbol`s or `str`s that can be parsed into a `clingo.symbol.Symbol`s with `clingo.symbol.parse_term`.
    """

    def __init__(self, symbols: Set[Union[Symbol, str]]) -> None:
        self.__symbols = {_into_symbol(s) for s in symbols}

    def __repr__(self):
        name = self.__class__.__name__
        symbols = {str(symbol) for symbol in self.__symbols}
        return f"{name}({symbols})"

    def holds_for(self, model: Model) -> bool:
        return set(model.symbols(shown=True)).issuperset(self.__symbols)


class True_(Assertion):
    """
    The assertion that is true for each model.
    """

    def __repr__(self):
        return f"{self.__class__.__name__}()"

    def holds_for(self, model: Model) -> bool:
        return True


class False_(Assertion):
    """
    The assertion that is false for each model.
    """

    def __repr__(self):
        return f"{self.__class__.__name__}()"

    def holds_for(self, model: Model) -> bool:
        return False


class Not(Assertion):
    """
    The negation of a given assertion.
    This assertion holds if `operand` does not hold and vice versa.

    Parameters
    ----------
    operand
        The `Assertion` to be negated.
    """

    def __init__(self, operand: Assertion) -> None:
        self.__operand = operand

    def __repr__(self):
        name = self.__class__.__name__
        operand = repr(self.__operand)
        return f"{name}({operand})"

    def holds_for(self, model: Model) -> bool:
        return not self.__operand.holds_for(model)


class And(Assertion):
    """
    The conjunction of a list of given assertions.
    This assertion holds if all `args` hold.

    Parameters
    ----------
    args
        The `Assertion`s to be combined.
    """

    def __init__(self, *args: Assertion) -> None:
        self.__operands = args

    def __repr__(self):
        name = self.__class__.__name__
        operands = ", ".join(repr(operand) for operand in self.__operands)
        return f"{name}({operands})"

    def holds_for(self, model: Model) -> bool:
        return all((operand.holds_for(model) for operand in self.__operands))


class Or(Assertion):
    """
    The disjunction of a list of given assertions.
    This assertions hold if any `args` hold.

    Parameters
    ----------
    args
        The `Assertion`s to be combined.
    """

    def __init__(self, *args: Assertion) -> None:
        self.__operands = args

    def __repr__(self):
        name = self.__class__.__name__
        operands = ", ".join(repr(operand) for operand in self.__operands)
        return f"{name}({operands})"

    def holds_for(self, model: Model) -> bool:
        return any((operand.holds_for(model) for operand in self.__operands))



# pylint: disable=import-outside-toplevel
# pylint: disable=missing-module-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring
# pylint: disable=duplicate-code


def test_clingo():
    from clintest.solver import Clingo
    from clintest.test import Record, Recording

    solver = Clingo("0", "a.")
    test = Record()

    solver.solve(test)
    assert Recording([
        {'__f': '__init__'},
        {'__f': 'on_model', 'str(model)': 'a'},
        {'__f': 'on_statistics'},
        {'__f': 'on_finish'},
    ]).subsumes(test.recording)



"""
The abstract class `clintest.test.Test` and off-the-shelf test implementations.
"""

from abc import ABC, abstractmethod
import os
from textwrap import indent
from typing import Any, Callable, Dict, Optional, Sequence

from clingo.solving import Model, SolveResult
from clingo.statistics import StatisticsMap

from .outcome import Outcome
from .quantifier import Quantifier, Finished
from .assertion import Assertion


class Test(ABC):
    """
    An abstract test consuming the artifacts of a `clintest.solver.Solver` in order to compute an
    `clintest.outcome.Outcome`.
    """

    def on_model(self, _model: Model) -> bool:
        """
        Consume a `clingo.model.Model` and possibly alter the current outcome of this test.

        Parameters
        ----------
        model
            The `clingo.model.Model` to consume.

        Returns
        -------
        Whether further models a needed to decide this test.
        """

        return True

    def on_unsat(self, lower_bound: Sequence[int]) -> None:
        """
        Consume a `lower_bound` during optimization and possibly alter the current outcome of this
        test.

        Parameters
        ----------
        lower_bound
            The lower bound.
        """

    def on_core(self, core: Sequence[int]) -> None:
        """
        Consume an unsat `core` and possibly alter the current outcome of this test.

        Parameters
        ----------
        core
            The unsat core.
        """

    def on_statistics(self, step: StatisticsMap, accumulated: StatisticsMap) -> None:
        """
        Consume the solving statistics and possibly alter the current outcome of this test.

        Parameters
        ----------
        step
            The step statistics.

        accumulated
            The accumulated statistics.
        """

    @abstractmethod
    def on_finish(self, result: SolveResult) -> None:
        """
        Consume the final solve result and possibly alter the current outcome of this test.

        This should be the last `on_*`-method ever called on a test.
        Afterwards the outcome must be certain.

        Parameters
        ----------
        result
            The `clingo.solving.SolveResult`.
        """

    @abstractmethod
    def outcome(self) -> Outcome:
        """
        Returns the current `Outcome` of this test.

        Returns
        -------
        The current outcome of this test.
        """

    def assert_(self) -> None:
        """
        Assert the outcome of this test to be certainly true.
        Raise an `AssertionError` if the test is either incomplete or has failed.
        """

        if not self.outcome().is_certainly_true():
            msg = "The following test "
            msg += ["is incomplete.", "has failed."][self.outcome().is_certain()]
            msg += os.linesep
            msg += indent(str(self), 4 * " ")

            raise AssertionError(msg)


class True_(Test):
    """
    The test which always succeeds.

    Parameters
    ----------
    lazy
        Whether this test should be lazy, i.e., not consume any models.
    """

    def __init__(self, lazy: bool = True) -> None:
        self.__outcome = Outcome(True, lazy)

    def __repr__(self):
        name = self.__class__.__name__
        outcome = repr(self.__outcome)
        return f"{name}(__outcome={outcome})"

    def __str__(self):
        return f"[{self.__outcome}] {self.__class__.__name__}"

    def on_model(self, _model: Model) -> bool:
        return not self.__outcome.is_certain()

    def on_finish(self, result: SolveResult) -> None:
        self.__outcome = Outcome(True, True)

    def outcome(self) -> Outcome:
        return self.__outcome


class False_(Test):
    """
    The test which always failes.

    Parameters
    ----------
    lazy
        Whether this test should be lazy, i.e., not consume any models.
    """

    def __init__(self, lazy: bool = True) -> None:
        self.__outcome = Outcome(False, lazy)

    def __repr__(self):
        name = self.__class__.__name__
        outcome = repr(self.__outcome)
        return f"{name}(__outcome={outcome})"

    def __str__(self):
        return f"[{self.__outcome}] {self.__class__.__name__}"

    def on_model(self, _model: Model) -> bool:
        return not self.__outcome.is_certain()

    def on_finish(self, result: SolveResult) -> None:
        self.__outcome = Outcome(False, True)

    def outcome(self) -> Outcome:
        return self.__outcome


class Recording:
    """
    A recording of the calls to the `on_*`-methods of a `Test`.
    This class is mainly used inside of `Record`.
    """

    def __init__(self, entries: Optional[Sequence[Dict[str, Any]]] = None):
        if entries is None:
            entries = []
        self.__entries = list(entries)

    def __repr__(self):
        name = self.__class__.__name__
        return f"{name}({self.__entries})"

    def __str__(self):
        def fmt(entry):
            result = f"[{entry['__outcome']}] {entry['__f']}"
            if entry['__f'] == "on_model":
                result += os.linesep + 4 * " " + entry['str(model)']
            return result

        width = len(str(len(self.__entries) - 1))
        return os.linesep.join((
            f"{(width - len(str(i))) * ' '}{i}: {fmt(entry)}"
            for i, entry in enumerate(self.__entries)
        ))

    def __eq__(self, other):
        # pylint: disable=protected-access
        return self.__entries == other.__entries

    def amend(self, changes: Dict[str, Any]):
        """
        Update the last entry of this encoding.

        Parameters
        ----------
        changes
            The changes.
        """

        self.__entries[-1].update(changes)

    def append(self, entry: Dict[str, Any]):
        """
        Append a new entry at the end of this recording.

        Parameters
        ----------
        entry
            The entry.
        """

        self.__entries.append(entry)

    def subsumes(self, other) -> bool:
        """
        Determine whether this recording subsumes another recording.

        Parameters
        ----------
        other
            The other recording.
        """

        # pylint: disable=protected-access
        return len(self.__entries) == len(other.__entries) and all(
            all(item in other_entry.items() for item in self_entry.items())
            for self_entry, other_entry in zip(self.__entries, other.__entries)
        )


class Record(Test):
    """
    A test that behaves identical to a given other `test` but records any call to one of it
    `on_*`-methods. This can be very helpful for debugging.

    Parameters
    ----------
    test
        A `Test` that determines how this test should behave.
    """

    def __init__(self, test: Test = True_(lazy = False)):
        self.test: Test = test
        self.recording: Recording = Recording([{
            "__f": "__init__",
            "__outcome": self.outcome(),
        }])

    def __repr__(self):
        name = self.__class__.__name__
        test = repr(self.test)
        recording = repr(self.recording)
        return f"{name}(test={test}, recording={recording})"

    def __str__(self):
        return os.linesep.join([
            f"[{self.outcome()}] {self.__class__.__name__}",
            f"    test: {indent(str(self.test), 4 * ' ')[4:]}",
            "    recording:",
            indent(str(self.recording), 8 * " "),
        ])

    def on_model(self, model: Model) -> bool:
        self.recording.append({
            "__f": "on_model",
            "str(model)": str(model),
        })
        result = self.test.on_model(model)
        self.recording.amend({
            "__result": result,
            "__outcome": self.outcome(),
        })
        return result

    def on_unsat(self, lower_bound: Sequence[int]) -> None:
        self.recording.append({
            "__f": "on_unsat",
            "lower_bound": lower_bound,
        })
        self.test.on_unsat(lower_bound)
        self.recording.amend({"__outcome": self.outcome()})

    def on_core(self, core: Sequence[int]) -> None:
        self.recording.append({
            "__f": "on_core",
            "core": core,
        })
        self.test.on_core(core)
        self.recording.amend({"__outcome": self.outcome()})

    def on_statistics(self, step: StatisticsMap, accumulated: StatisticsMap) -> None:
        self.recording.append({
            "__f": "on_statistics",
            "step": step,
            "accumulated": accumulated,
        })
        self.test.on_statistics(step, accumulated)
        self.recording.amend({"__outcome": self.outcome()})

    def on_finish(self, result: SolveResult) -> None:
        self.recording.append({
            "__f": "on_finish",
            "result": result,
        })
        self.test.on_finish(result)
        self.recording.amend({"__outcome": self.outcome()})

    def outcome(self) -> Outcome:
        return self.test.outcome()


class Assert(Test):
    """
    A test that asserts certain properties about the `clingo.model.Model`s of a program. This test
    can be highly costumized using a `clintest.quantifier.Quantifier` and a
    `clintest.assertion.Assertion`.

    Parameters
    ----------
    quantifier
        The `clintest.quantifier.Quantifier` used with this test.

    assertion
        The `clintest.assertion.Assertion` used with this test.
    """

    def __init__(self, quantifier: Quantifier, assertion: Assertion) -> None:
        self.__quantifier = quantifier
        self.__assertion = assertion

    def __repr__(self):
        name = self.__class__.__name__
        quantifier = repr(self.__quantifier)
        assertion = repr(self.__assertion)
        return f"{name}({quantifier}, {assertion})"

    def __str__(self):
        return os.linesep.join([
            f"[{self.outcome()}] {self.__class__.__name__}",
            f"    quantifier: {self.__quantifier}",
            f"    assertion:  {self.__assertion}",
        ])

    def on_model(self, model: Model) -> bool:
        if not self.__quantifier.outcome().is_certain():
            self.__quantifier.consume(self.__assertion.holds_for(model))

        return not self.__quantifier.outcome().is_certain()

    def on_finish(self, result: SolveResult) -> None:
        self.__quantifier = Finished(self.__quantifier)

    def outcome(self) -> Outcome:
        return self.__quantifier.outcome()


class Not(Test):
    """
    The negation of a given test.
    This test failes if `operand` succeeds and vice versa.

    Parameters
    ----------
    operand
        The `Test` to be negated.
    """

    def __init__(self, operand: Test) -> None:
        self.__operand = operand

    def __repr__(self):
        name = self.__class__.__name__
        operand = repr(self.__operand)
        return f"{name}({operand})"

    def __str__(self):
        return os.linesep.join([
            f"[{self.outcome()}] {self.__class__.__name__}",
            f"    operand: {indent(str(self.__operand), 4 * ' ')[4:]}",
        ])

    def on_model(self, model: Model) -> bool:
        return self.__operand.on_model(model)

    def on_unsat(self, lower_bound: Sequence[int]) -> None:
        self.__operand.on_unsat(lower_bound)

    def on_core(self, core: Sequence[int]) -> None:
        self.__operand.on_core(core)

    def on_statistics(self, step: StatisticsMap, accumulated: StatisticsMap) -> None:
        self.__operand.on_statistics(step, accumulated)

    def on_finish(self, result: SolveResult) -> None:
        self.__operand.on_finish(result)

    def outcome(self) -> Outcome:
        outcome = self.__operand.outcome()
        return Outcome(not outcome.current_value(), outcome.is_certain())


class And(Test):
    """
    The conjunction of a list given tests.
    This test succeeds if all `args` succeed.

    Parameters
    ----------
    args
        The `Test`s to be combined.

    short_circuit
        Whether this test should employ short circuit optimization, i.e., abort all remaining tests
        once the outcome of a test is certainly false.

    ignore_certain
        Whether this test should employ the ignore certain optimization, i.e., not send artifacts
        to test that are already certain.
    """

    def __init__(
        self,
        *args: Test,
        short_circuit: bool = True,
        ignore_certain: bool = True
    ) -> None:
        self.__operands = list(args)
        self.__short_circuit = short_circuit
        self.__ignore_certain = ignore_certain

        self.__ongoing = list(args)
        self.__outcome = Outcome(True, False)

        def call_operand(_operand: Test) -> None:
            pass

        self.__on_whatever(call_operand)

    def __repr__(self):
        name = self.__class__.__name__

        operands = ", ".join(repr(operand) for operand in self.__operands)
        short_circuit = repr(self.__short_circuit)
        ignore_certain = repr(self.__ignore_certain)

        ongoing = repr(self.__ongoing)
        outcome = repr(self.__outcome)

        return (
            f"{name}("
            f"{operands}, "
            f"short_circuit={short_circuit}, "
            f"ignore_certain={ignore_certain}, "
            f"__ongoing={ongoing}, "
            f"__outcome={outcome})"
        )

    def __str__(self):
        if self.__operands:
            operands = ""
            width = len(str(len(self.__operands) - 1))
            for i, operand in enumerate(self.__operands):
                i = str(i)
                operands += os.linesep
                operands += (width - len(i)) * " "
                operands += [" ", "*"][operand in self.__ongoing]
                operands += f"{i}: {operand}"
            operands = indent(operands, 8 * ' ')
        else:
            operands = " <none>"

        return os.linesep.join([
            f"[{self.outcome()}] {self.__class__.__name__} ",
            f"    operands:{operands}",
            f"    short_circuit:  {self.__short_circuit}",
            f"    ignore_certain: {self.__ignore_certain}",
        ])

    def __on_whatever(self, call_operand: Callable[[Test], None]) -> bool:
        still_ongoing = []

        for operand in self.__ongoing:
            call_operand(operand)

            if operand.outcome().is_certainly_false():
                if self.__short_circuit:
                    self.__ongoing = []
                    self.__outcome = Outcome(False, True)
                    return False
                else:
                    self.__outcome = Outcome(False, False)

            if not (self.__ignore_certain and operand.outcome().is_certain()):
                still_ongoing.append(operand)

        self.__ongoing = still_ongoing
        self.__outcome = Outcome(self.__outcome.current_value(), not bool(still_ongoing))

        return not self.__outcome.is_certain()

    def on_model(self, model: Model) -> bool:
        def call_operand(operand: Test) -> None:
            operand.on_model(model)

        return self.__on_whatever(call_operand)

    def on_unsat(self, lower_bound: Sequence[int]) -> None:
        def call_operand(operand: Test) -> None:
            operand.on_unsat(lower_bound)

        self.__on_whatever(call_operand)

    def on_core(self, core: Sequence[int]) -> None:
        def call_operand(operand: Test) -> None:
            operand.on_core(core)

        self.__on_whatever(call_operand)

    def on_statistics(self, step: StatisticsMap, accumulated: StatisticsMap) -> None:
        def call_operand(operand: Test) -> None:
            operand.on_statistics(step, accumulated)

        self.__on_whatever(call_operand)

    def on_finish(self, result: SolveResult) -> None:
        def call_operand(operand: Test) -> None:
            operand.on_finish(result)

        ignore_certain_bck = self.__ignore_certain
        self.__ignore_certain = True
        self.__on_whatever(call_operand)
        self.__ignore_certain = ignore_certain_bck

        assert not self.__ongoing
        assert self.__outcome.is_certain()

    def outcome(self) -> Outcome:
        return self.__outcome


class Or(Test):
    """
    The disjunction of a list given tests.
    This test succeeds if any `args` succeed.

    Parameters
    ----------
    args
        The `Test`s to be combined.

    short_circuit
        Whether this test should employ short circuit optimization, i.e., abort all remaining tests
        once the outcome of a test is certainly true.

    ignore_certain
        Whether this test should employ the ignore certain optimization, i.e., not send artifacts
        to test that are already certain.
    """

    def __init__(
        self,
        *args: Test,
        short_circuit: bool = True,
        ignore_certain: bool = True
    ) -> None:
        self.__operands = list(args)
        self.__short_circuit = short_circuit
        self.__ignore_certain = ignore_certain

        self.__ongoing = list(args)
        self.__outcome = Outcome(False, False)

        def call_operand(_operand: Test) -> None:
            pass

        self.__on_whatever(call_operand)

    def __repr__(self):
        name = self.__class__.__name__

        operands = ", ".join(repr(operand) for operand in self.__operands)
        short_circuit = repr(self.__short_circuit)
        ignore_certain = repr(self.__ignore_certain)

        ongoing = repr(self.__ongoing)
        outcome = repr(self.__outcome)

        return (
            f"{name}("
            f"{operands}, "
            f"short_circuit={short_circuit}, "
            f"ignore_certain={ignore_certain}, "
            f"__ongoing={ongoing}, "
            f"__outcome={outcome})"
        )

    def __str__(self):
        if self.__operands:
            operands = ""
            width = len(str(len(self.__operands) - 1))
            for i, operand in enumerate(self.__operands):
                i = str(i)
                operands += os.linesep
                operands += (width - len(i)) * " "
                operands += [" ", "*"][operand in self.__ongoing]
                operands += f"{i}: {operand}"
            operands = indent(operands, 8 * ' ')
        else:
            operands = " <none>"

        return os.linesep.join([
            f"[{self.outcome()}] {self.__class__.__name__} ",
            f"    operands:{operands}",
            f"    short_circuit:  {self.__short_circuit}",
            f"    ignore_certain: {self.__ignore_certain}",
        ])

    def __on_whatever(self, call_operand: Callable[[Test], None]) -> bool:
        still_ongoing = []

        for operand in self.__ongoing:
            call_operand(operand)

            if operand.outcome().is_certainly_true():
                if self.__short_circuit:
                    self.__ongoing = []
                    self.__outcome = Outcome(True, True)
                    return False
                else:
                    self.__outcome = Outcome(True, False)

            if not (self.__ignore_certain and operand.outcome().is_certain()):
                still_ongoing.append(operand)

        self.__ongoing = still_ongoing
        self.__outcome = Outcome(self.__outcome.current_value(), not bool(still_ongoing))

        return not self.__outcome.is_certain()

    def on_model(self, model: Model) -> bool:
        def call_operand(operand: Test) -> None:
            operand.on_model(model)

        return self.__on_whatever(call_operand)

    def on_unsat(self, lower_bound: Sequence[int]) -> None:
        def call_operand(operand: Test) -> None:
            operand.on_unsat(lower_bound)

        self.__on_whatever(call_operand)

    def on_core(self, core: Sequence[int]) -> None:
        def call_operand(operand: Test) -> None:
            operand.on_core(core)

        self.__on_whatever(call_operand)

    def on_statistics(self, step: StatisticsMap, accumulated: StatisticsMap) -> None:
        def call_operand(operand: Test) -> None:
            operand.on_statistics(step, accumulated)

        self.__on_whatever(call_operand)

    def on_finish(self, result: SolveResult) -> None:
        def call_operand(operand: Test) -> None:
            operand.on_finish(result)

        ignore_certain_bck = self.__ignore_certain
        self.__ignore_certain = True
        self.__on_whatever(call_operand)
        self.__ignore_certain = ignore_certain_bck

        assert not self.__ongoing
        assert self.__outcome.is_certain()

    def outcome(self) -> Outcome:
        return self.__outcome



# pylint: disable=import-outside-toplevel
# pylint: disable=missing-module-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring
# pylint: disable=redefined-builtin


def test_all():
    from clintest.quantifier import All
    quantifier = All()

    assert quantifier.outcome().as_tuple() == (True, False)

    input = 2 * [True, False]
    output = [quantifier.consume(value).as_tuple() for value in input]

    assert output == [
        (True, False),
        (False, True),
        (False, True),
        (False, True),
    ]


def test_any():
    from clintest.quantifier import Any
    quantifier = Any()

    assert quantifier.outcome().as_tuple() == (False, False)

    input = 2 * [False, True]
    output = [quantifier.consume(value).as_tuple() for value in input]

    assert output == [
        (False, False),
        (True, True),
        (True, True),
        (True, True),
    ]

def test_exact():
    from clintest.quantifier import Exact
    quantifier = Exact(2)

    assert quantifier.outcome().as_tuple() == (False, False)

    input = 4 * [False, True]
    output = [quantifier.consume(value).as_tuple() for value in input]

    assert output == [
        (False, False),
        (False, False),
        (False, False),
        (True, False),
        (True, False),
        (False, True),
        (False, True),
        (False, True),
    ]



# pylint: disable=import-outside-toplevel
# pylint: disable=missing-module-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring
# pylint: disable=redefined-outer-name


from clingo import Control
import pytest


@pytest.fixture
def frame():
    ctl = Control(["0"])
    ctl.add("base", [], "a. b.")
    ctl.ground([("base", [])])

    def test(positive, negative):
        for model in ctl.solve(yield_=True):
            for assertion in positive:
                assert     assertion.holds_for(model)
            for assertion in negative:
                assert not assertion.holds_for(model)

    return test


def test_contains(frame):
    from clintest.assertion import Contains
    frame([
        Contains("a"),
        Contains("b"),
    ], [
        Contains("c"),
    ])


def test_equals(frame):
    from clintest.assertion import Equals
    frame([
        Equals({"a", "b"}),
    ], [
        Equals({"a"}),
        Equals({"a", "b", "c"}),
    ])


def test_subsetof(frame):
    from clintest.assertion import SubsetOf
    frame([
        SubsetOf({"a", "b"}),
        SubsetOf({"a", "b", "c"}),
    ], {
        SubsetOf({"a"}),
    })


def test_supersetof(frame):
    from clintest.assertion import SupersetOf
    frame([
        SupersetOf({"a"}),
        SupersetOf({"a", "b"}),
    ], {
        SupersetOf({"a", "b", "c"}),
    })


def test_true(frame):
    from clintest.assertion import True_
    frame([True_()], [])


def test_false(frame):
    from clintest.assertion import False_
    frame([], [False_()])


def test_not(frame):
    from clintest.assertion import Not, True_, False_
    frame([Not(False_())], [Not(True_())])


def test_and(frame):
    from clintest.assertion import And, True_, False_
    frame([
        And(),
        And(True_(), True_()),
    ], [
        And(True_(), False_()),
        And(False_(), True_()),
        And(False_(), False_()),
    ])


def test_or(frame):
    from clintest.assertion import Or, True_, False_
    frame([
        Or(True_(), True_()),
        Or(True_(), False_()),
        Or(False_(), True_()),
    ], [
        Or(),
        Or(False_(), False_()),
    ])



"""
The abstract class `clintest.quantifier.Quantifier` and classes extending it.
"""

from abc import ABC, abstractmethod

from .outcome import Outcome


class Quantifier(ABC):
    """
    A quantifier specifies how many assertions must hold in order to pass the test.
    As such, one is necessary to assemble the `clintest.test.Assert` test.

    Quantifiers are stateful.
    They consume the return values of `clintest.assertion.Assertion.holds_for` and store all the
    information necessary to determine the current outcome of the test.
    """

    @abstractmethod
    def outcome(self) -> Outcome:
        """
        Returns the current outcome of this quantifier.
        """

    @abstractmethod
    def consume(self, value: bool) -> Outcome:
        """
        Consume the return value of `clintest.assertion.Assertion.holds_for` and possibly alter the
        current outcome of this quantifier.

        Parameters
        ----------
        value
            The return value of `clintest.assertion.Assertion.holds_for`.

        Returns
        -------
        The outcome of this quantifier after `value` was consumed.
        """


class All(Quantifier):
    """
    A quantifier demanding that an assertion holds for all models.
    """

    def __init__(self) -> None:
        self.__state = Outcome(True, False)

    def __repr__(self):
        name = self.__class__.__name__
        state = repr(self.__state)
        return f"{name}(__state={state})"

    def __str__(self):
        return self.__class__.__name__

    def outcome(self) -> Outcome:
        return self.__state

    def consume(self, value: bool) -> Outcome:
        if not value:
            self.__state = Outcome(False, True)
        return self.__state


class Any(Quantifier):
    """
    A quantifier demanding that an assertion holds for any model.
    """

    def __init__(self) -> None:
        self.__state = Outcome(False, False)

    def __repr__(self):
        name = self.__class__.__name__
        state = repr(self.__state)
        return f"{name}(__state={state})"

    def __str__(self):
        return self.__class__.__name__

    def outcome(self) -> Outcome:
        return self.__state

    def consume(self, value: bool) -> Outcome:
        if value:
            self.__state = Outcome(True, True)
        return self.__state


class Exact(Quantifier):
    """
    A quantifier demanding that an assertion holds for an exact number of models.

    Parameters
    ----------
    target
        The number of models the assertion should hold for
    """

    def __init__(self, target: int) -> None:
        self.__target = target
        self.__state = 0

    def __repr__(self):
        name = self.__class__.__name__
        return f"{name}({self.__target}, __state={self.__state})"

    def __str__(self):
        return f"{self.__class__.__name__} {self.__state}/{self.__target}"

    def outcome(self) -> Outcome:
        return Outcome(self.__state == self.__target, self.__state > self.__target)

    def consume(self, value: bool) -> Outcome:
        self.__state += value
        return self.outcome()


class Finished(Quantifier):
    """
    A wrapper around an `inner` quantifier indicating that computation has finished.
    The outcome of this quantifier is the outcome of `inner` beside that it is always certain.
    Calling `Finished.consume` will not alter outcome of this or the `inner` quantifier.

    Parameters
    ---------
    inner
        The quantifier that should be finished.
    """

    def __init__(self, inner: Quantifier) -> None:
        self.__state = Outcome(inner.outcome().current_value(), True)

    def __repr__(self):
        name = self.__class__.__name__
        state = repr(self.__state)
        return f"{name}(__state={state})"

    def __str__(self):
        return self.__class__.__name__

    def outcome(self) -> Outcome:
        return self.__state

    def consume(self, value: bool) -> Outcome:
        return self.__state



"""
The abstract class `clintest.solver.Solver` and off-the-shelf solver implementations.
"""

from abc import ABC, abstractmethod
from typing import Optional, Sequence

from clingo.control import Control

from .test import Test


class Solver(ABC):
    """
    An initialized solver that may solve any test.
    """

    @abstractmethod
    def solve(self, test: Test) -> None:
        """
        Use this solver to solve a given `test`.

        Parameters
        ----------
        test
            The `clintest.test.Test` to be solved by this solver.
        """


class Clingo(Solver):
    """
    A solver using `clingo.control.Control`.

    Parameters
    ----------
    arguments
        A list of arguments.

    program
        The program as a `str`.

    files
        A list of files to read the program from.
    """

    def __init__(
        self,
        arguments: Optional[Sequence[str]] = None,
        program: Optional[str] = None,
        files: Optional[Sequence[str]] = None,
    ) -> None:
        self.__arguments = [] if arguments is None else arguments
        self.__program = "" if program is None else program
        self.__files = [] if files is None else files

    def solve(self, test: Test) -> None:
        ctl = Control(self.__arguments)

        ctl.add("base", [], self.__program)

        for file in self.__files:
            ctl.load(file)

        ctl.ground([("base", [])])

        if not test.outcome().is_certain():
            ctl.solve(
                on_model=test.on_model,
                on_unsat=test.on_unsat,
                on_core=test.on_core,
                on_statistics=test.on_statistics,
                on_finish=test.on_finish,
            )

    def __repr__(self):
        name = self.__class__.__name__
        arguments = repr(self.__arguments)
        program = repr(self.__program)
        files = repr(self.__files)
        return f"{name}({arguments}, {program}, {files})"
